//
// fldx.cpp
// 
// FLTK windowing code for dCam capture application
//
//
// Copyright 2000 by Videre Design 
//


#include "dcam.h"
#include "timer.h"
#include "image_io.h"
#ifdef WIN32
#include <windows.h>
#include <mmsystem.h>
#include "direct.h"
#else
#include <unistd.h>
#define Sleep(x) usleep(1000*x)
#include <ctype.h>
#endif
#include <stdlib.h>
#include <stdarg.h>

#include "fld.h"
#include "FL/Fl_File_Chooser.H"
#include "FL/Fl_Browser.H"

// set/unset this to draw an overlay rectangle demo
static bool drawOverlayFlag = false;

// debugging window
dcamDebugWin *debugWin = NULL;	// for debugging output

// variables
static fld *mainw;		// main window class instance
static dCamera *dcam = NULL;	// current camera
static int cam_num = 0;		// selected camera
static bool single_shot = false; // for single-shot mode
static dITYPE cam_type = S_RGB24; // current values
static dISPEED cam_speed = S_15;
static dISIZE cam_size = S_640x480;
static int cam_fdiv = 1;

// acquisition mode
#define AQT_NONE     0
#define AQT_CONTIN   1
#define AQT_BUFFER   2
int acquire_type = AQT_NONE;

// acquisition params
unsigned char *this_image = NULL;	// global holds image data
int im_width, im_height, im_color, im_size; // image params


// buffer list
#define MAX_BUFFER_FRAMES 200
int bufferNum = 0;              // current number of frames in buffer
int bufferCur = 0;		// current frame
int bufferSize = MAX_BUFFER_FRAMES; // size of buffer
unsigned char *bufferList[MAX_BUFFER_FRAMES];
bool storeFlag = false;


//
// info and debug window messages
//

void debugFn(char *str)
{
  if (debugWin)
    {
      if (str)
	debugWin->Print(str);
    }
}

void
status_message(char *s)
{
  if (mainw)
    mainw->infoLine->value(s);
}


void
debugMessage(char *str, ...)
{
  char buf[1024];
  va_list ptr;
  va_start(ptr,str);
  vsprintf(buf,str,ptr);
  status_message(buf);
  debugFn(buf);
  va_end(ptr);
}




// set GUI to match info

void                    // sets choice value for this button
show_format()
{
}

void
show_size()
{
  //    mainw->size_val->value(1);
  show_format();
}


void
show_dig_params()
{
  if (dcam)
    {
      mainw->exposure_val->value(dcam->exposure);
      mainw->gain_val->value(dcam->gain);
    }
}

void
show_all()
{
  show_size();
  show_dig_params();
}




// destructor for fldisp, ends the session

fld::~fld()
{
  // do we need any teardown here?
  exit(0);
}


//
// Here we do the callbacks for all buttons and valuators
//

// exit callback, just end the session

void
exit_cb(Fl_Menu_ *w, void *u)
{
  delete mainw;
}


// Input callback: open or close video, or use buffer

void
input_cb(Fl_Choice *w, void *u)
{
  Fl_Menu_ *mw = (Fl_Menu_ *)w;
  const Fl_Menu_Item* m = mw->mvalue();

  if (!strcmp(m->label(), "None"))
    {
      acquire_type = AQT_NONE;
      if (dcam)
        dcam->Stop();
      single_shot = false;
      mainw->contin_val->value(false);
      mainw->contin_val->deactivate();
//      mainw->single_val->deactivate();
      mainw->record->deactivate();
      mainw->clear->deactivate();
      mainw->frame_val->deactivate();
      mainw->camera_val->activate();
#ifdef DCAM_FW_DIALOG
      mainw->firmware_params_val->deactivate();
      mainw->menubar->redraw();
#endif
    }

  if (!strcmp(m->label(), "Video"))
    {
      acquire_type = AQT_NONE;
      dcam = dSys->InitCamera(cam_num);
      if (dcam == NULL)
        {
          status_message("Can't initialize camera");
          mainw->input_val->value(0);
          return;
        }
      single_shot = false;
      status_message("Camera initialized");
      show_all();
      mainw->contin_val->activate();
//      mainw->single_val->activate();
      mainw->clear->activate();
      mainw->record->activate();
      mainw->frame_val->activate();
      mainw->format_val->activate();
      mainw->speed_val->activate();
      mainw->camera_val->deactivate(); // doesn't work...

      // set color/mono
      if (dcam->has_color)
	{
	  mainw->format_val->value(0);
	  cam_type = S_RGB24;
	}
      else
	{
	  mainw->format_val->value(1);
	  cam_type = S_MONO;
	}

      // set frame div 
      if (dcam->has_frame_div)
	{
	  mainw->fdiv_val->activate();
	  Fl_Menu_Item *m = (Fl_Menu_Item *)mainw->fdiv_val->menu();
	}
      else
	mainw->fdiv_val->deactivate();

      // set video parameters
      if (dcam->has_auto_whitebalance)
	{
	  mainw->balance_auto_button->value(true);
	  mainw->balance_auto_button->activate();
	  mainw->red_val->deactivate();
	  mainw->blue_val->deactivate();
	}
      else if (dcam->has_color)
	{
	  mainw->balance_auto_button->value(false);
	  mainw->balance_auto_button->deactivate();
	  mainw->red_val->activate();
	  mainw->blue_val->activate();
	}
      else
	{
	  mainw->balance_auto_button->value(false);
	  mainw->balance_auto_button->deactivate();
	  mainw->red_val->deactivate();
	  mainw->blue_val->deactivate();
	}

      if (dcam->has_auto_brightness)
	{
	  mainw->brightness_auto_button->value(true);
	  mainw->brightness_auto_button->activate();
	  mainw->brightness_val->deactivate();
	}
      else if (dcam->has_manual_brightness)
	{
	  mainw->brightness_auto_button->value(false);
	  mainw->brightness_auto_button->deactivate();
	  mainw->brightness_val->activate();
	}
      else
	{
	  mainw->brightness_auto_button->value(false);
	  mainw->brightness_auto_button->deactivate();
	  mainw->brightness_val->deactivate();
	}

      if (dcam->has_auto_exposure)
	{
	  mainw->exposure_auto_button->value(true);
	  mainw->exposure_auto_button->activate();
	  mainw->exposure_val->deactivate();
	}
      else if (dcam->has_manual_exposure)
	{
	  mainw->exposure_auto_button->value(false);
	  mainw->exposure_auto_button->deactivate();
	  mainw->exposure_val->activate();
	  mainw->gain_val->activate();
	}
      else
	{
	  mainw->exposure_auto_button->value(false);
	  mainw->exposure_auto_button->deactivate();
	  mainw->exposure_val->deactivate();
	  mainw->gain_val->deactivate();
	}

      if (dcam->has_auto_gain)
	{
	  mainw->gain_auto_button->value(false);
	  mainw->gain_auto_button->activate();
	  mainw->gain_val->activate();
	}
      else if (dcam->has_manual_gain)
	{
	  mainw->gain_auto_button->value(false);
	  mainw->gain_auto_button->deactivate();
	  mainw->gain_val->activate();
	}
      else
	{
	  mainw->gain_auto_button->value(false);
	  mainw->gain_auto_button->deactivate();
	  mainw->gain_val->deactivate();
	}

      // saturation
      if (dcam->has_saturation)
	mainw->saturation_val->activate();
      else
	mainw->saturation_val->deactivate();

      // sharpness
      if (dcam->has_sharpness)
	mainw->sharpness_val->activate();
      else
	mainw->sharpness_val->deactivate();


      // check for gamma on display
      if (!dcam->has_gamma || dcam->isBayer)
        mainw->image->displayGamma = 0.75;
      else
        mainw->image->displayGamma = 1.0;
      mainw->display_gamma_val->value(mainw->image->displayGamma * 100);

      // check for gamma on imager
      if (dcam->has_gamma)
	{
	  mainw->imager_gamma_button->activate();
	  if (dcam->gamma > 0)
	    mainw->imager_gamma_button->value(true);
	  else
	    mainw->imager_gamma_button->value(false);
	}
      else
	{
	  mainw->imager_gamma_button->deactivate();
	  mainw->imager_gamma_button->value(false);
	}

      // set acquire mode flag
      acquire_type = AQT_CONTIN;

#ifdef DCAM_FW_DIALOG
      mainw->firmware_params_val->activate();
      mainw->menubar->redraw();
#endif
    }

  if (!strcmp(m->label(), "Buffer"))
    {
      acquire_type = AQT_NONE;
      if (dcam)
        dcam->Stop();
      mainw->contin_val->activate();
//      mainw->single_val->activate();
      mainw->record->deactivate();
      mainw->clear->deactivate();
      mainw->frame_val->activate();
      mainw->camera_val->activate();
      acquire_type = AQT_BUFFER;
#ifdef DCAM_FW_DIALOG
      mainw->firmware_params_val->deactivate();
      mainw->menubar->redraw();
#endif
    }
}

// firmware parameter window callback

#ifdef DCAM_FW_DIALOG
extern void firmware_params_dialog(dCamera *dcam);
#endif

void
firmware_params_cb(class Fl_Menu_ *w, void *x)
{
#ifdef DCAM_FW_DIALOG
  firmware_params_dialog(dcam);
#endif
}


// size callback

void
size_cb(Fl_Choice *w, void *u)
{
  Fl_Menu_ *mw = (Fl_Menu_ *)w;
  const Fl_Menu_Item* m = mw->mvalue();

  if (!strcmp(m->label(), "160x120"))
    {
      cam_size = S_160x120;
    }
  if (!strcmp(m->label(), "320x240"))
    {
      cam_size = S_320x240;
    }

  if (!strcmp(m->label(), "640x480"))
    {
      cam_size = S_640x480;
    }
  if (!strcmp(m->label(), "1280x960"))
    {
      cam_size = S_1280x960;
    }
}


void
set_size()
{
  int old_size = im_size;

  // set up image parameters
  if (cam_size == S_1280x960)
    { im_width = 1280; im_height = 960; }
  if (cam_size == S_640x480)
    { im_width = 640; im_height = 480; }
  else if (cam_size == S_320x240)
    { im_width = 320; im_height = 240; }
  else if (cam_size == S_160x120)
    { im_width = 160; im_height = 120; }
  if (cam_type == S_MONO || cam_type == S_BAYER)
    {
      im_color = 0;
      im_size = im_width*im_height;
    }
  else
    {
      im_color = 1;
      im_size = im_width*im_height*3;
    }

  if (im_size != old_size)	// clear out buffer
    {
      int i;
      for (i=0; i<bufferSize; i++)
	{
	  if (bufferList[i])
	    delete [] bufferList[i];
	  bufferList[i] = NULL;
	}
    }
}


void
acquire_cb(Fl_Light_Button *w, void *u)
{
  if (!strcmp(w->label(),"Single"))
    {
      if (mainw->contin_val->value()) // continuous mode
        {
          debugMessage("Stop capture");
        }
      single_shot = true;
      mainw->contin_val->value(0);
      mainw->size_val->activate();
      mainw->format_val->activate();
      mainw->speed_val->activate();
      if (dcam && dcam->has_frame_div)
	mainw->fdiv_val->activate();
    }

  if (!strcmp(w->label(),"Continuous"))
    {
      if (mainw->contin_val->value()) // turn continuous mode on
        {
	  single_shot = false;
          bool ret = false;
          if (!dcam)
            debugMessage("No camera");
          else
            {
              ret = dcam->SetFormat(cam_size, cam_type, cam_speed, cam_fdiv);
              debugMessage("[Acquire] Video Format: %d %d %d", cam_size, cam_type, cam_speed);
              if (!ret)
                debugMessage("Video format: %s", dcam->error);
              else
                {
                  ret = dcam->Start();
                  if (!ret)
                    debugMessage("Can't get images");
                }
            }
          if (!ret)
            mainw->contin_val->value(false);
          else  
            {
	      set_size();
              debugMessage("Continuous capture");
              mainw->size_val->deactivate();
              mainw->format_val->deactivate();
              mainw->speed_val->deactivate();
	      mainw->fdiv_val->deactivate();
	      single_shot = true;
            }
        }
      else			// turn continuous mode off
        {
	  single_shot = false;
          if (dcam)
            dcam->Stop();
          debugMessage("Stop capture");
          mainw->size_val->activate();
          mainw->format_val->activate();
          mainw->speed_val->activate();
	  if (dcam && dcam->has_frame_div)
	    mainw->fdiv_val->activate();
        }
    }
}


// Video parameter callback

void
video_params_cb(Fl_Menu_ *w, void *u)
{
  if (!u) return;
  fld *flo = (fld *)u;
  Fl_Window *ww = flo->video_params;

  show_dig_params();
  ww->show();
}

// Debug window callback

void
debug_cb(Fl_Menu_ *w, void *u)
{
  debugWin->show();
}


// Reset imager callback

int off = 0;

void
reset_cb(Fl_Button *w, void *u)
{
  if (!dcam) return;
  dcam->Reset();
  //  off++;
  //  dcam->SetYOffset(off);
}

// Gain and exposure slider callbacks

void
set_dig_params()
{
  show_dig_params();
}

// Gain and exposure slider callbacks

void
set_exposure(void)
{
  if (!dcam) return;
  int val = (int)mainw->exposure_val->value();
  if (mainw->exposure_auto_button->value()) // come here on manual gain, auto exp
    dcam->SetExposure(val, true);
  else
    dcam->SetExposure(val, false);
}

void
set_gain(void)
{
  if (!dcam) return;
  int val = (int)mainw->gain_val->value();
  if (mainw->gain_auto_button->value()) // come here on manual gain, auto exp
    dcam->SetGain(val, true);
  else
    dcam->SetGain(val, false);
}

void
exposure_auto_cb(Fl_Light_Button *w, void *x)
{
  if (w->value())       // auto exposure
    {
      if (dcam)
	{
	  mainw->exposure_val->value(dcam->exposure);
	  mainw->exposure_val->deactivate();
	}
    }
  else
    {
      mainw->exposure_val->activate();
      if (dcam)
	{
	  mainw->exposure_val->value(dcam->exposure);
	}
      mainw->exposure_val->activate();
      show_dig_params();
    }
  set_exposure();
}

void
gain_auto_cb(Fl_Light_Button *w, void *x)
{
  if (w->value())       // auto gain
    {
      if (dcam)
	{
	  mainw->gain_val->value(dcam->gain);
	  mainw->gain_val->deactivate();
	}
    }
  else
    {
      mainw->gain_val->activate();
      if (dcam)
	{
	  mainw->gain_val->value(dcam->gain);
	}
      mainw->gain_val->activate();
      show_dig_params();
    }
  set_gain();
}


void
exposure_cb(Fl_Value_Slider *w, void *x)
{
  set_exposure();
}

void
gain_cb(Fl_Value_Slider *w, void *x)
{
  set_gain();
}


//
// Display gamma slider
//

void
display_gamma_cb(Fl_Slider *w, void *x)
{
  mainw->image->displayGamma = 0.01 * (double)(w->value());
}



//
// Color controls
//

void
set_balance(void)
{
  if (!dcam) return;
  int uval = (int)mainw->red_val->value();
  int vval = (int)mainw->blue_val->value();
  if (mainw->balance_auto_button->value())
    dcam->SetWhiteBalance(uval, vval, true);
  else
    dcam->SetWhiteBalance(uval, vval, false);
}

void
balance_auto_cb(Fl_Light_Button *w, void *x)
{
  if (w->value())       // auto exposure
    {
      mainw->red_val->deactivate();
      mainw->blue_val->deactivate();
    }
  else
    {
      mainw->red_val->activate();
      mainw->blue_val->activate();
    }

  set_balance();
}


void
red_gain_cb(Fl_Value_Slider *w, void *x)
{
  set_balance();
}

void
blue_gain_cb(Fl_Value_Slider *w, void *x)
{
  set_balance();
}

void
saturation_cb(Fl_Value_Slider *w, void *x)
{
  int sval = (int)mainw->saturation_val->value();
  if (!dcam) return;
  dcam->SetSaturation(sval);
}


// Brightness slider callbacks

void
set_brightness(void)
{
  if (!dcam) return;
  int bval = (int)mainw->brightness_val->value();
  if (mainw->brightness_auto_button->value())
    dcam->SetBrightness(bval, true);
  else
    dcam->SetBrightness(bval, false);
}

void
brightness_auto_cb(Fl_Light_Button *w, void *x)
{
  if (w->value() && (!dcam || dcam->has_auto_brightness)) // auto brightness
    mainw->brightness_val->deactivate();
  else
    {
      mainw->brightness_val->activate();
      mainw->brightness_auto_button->value(false);
    }
  set_brightness();
}


void
brightness_cb(Fl_Value_Slider *w, void *x)
{
  set_brightness();
}


// Sharpness, gamma callbacks

void
sharpness_cb(Fl_Value_Slider *w, void *x)
{
  if (!dcam) return;
  dcam->SetSharpness((int)w->value());
}

void
imager_gamma_cb(Fl_Light_Button *w, void *x)
{
  if (!dcam) return;
  dcam->SetGamma((bool)w->value());
}


// image format callback

void
format_val_cb(Fl_Choice *w, void *x)
{
  Fl_Menu_ *mw = (Fl_Menu_ *)w;
  const Fl_Menu_Item* m = mw->mvalue();

  if (!strcmp(m->label(), "RGB24"))
    {
      if (dcam && !dcam->has_color)
	{
	  mainw->format_val->value(1);
	  cam_type = S_MONO;
	}
      else
	{
	  cam_type = S_RGB24;
	  debugMessage("Camera type is RGB24");
	}
    }

  if (!strcmp(m->label(), "Mono"))
    {
      cam_type = S_MONO;
      debugMessage("Camera type is Monochrome");
    }

  if (!strcmp(m->label(), "Bayer"))
    {
      cam_type = S_BAYER;
//      cam_type = S_YUV422;
      debugMessage("Camera type is Bayer");
    }

}


// frame div callback
void
fdiv_val_cb(Fl_Choice *w, void *x)
{
  Fl_Menu_ *mw = (Fl_Menu_ *)w;
  const Fl_Menu_Item* m = mw->mvalue();

  if (!strcmp(m->label(), "1"))
    {
      cam_fdiv = 1;
    }

  if (!strcmp(m->label(), "2"))
    {
      cam_fdiv = 2;
    }

  if (!strcmp(m->label(), "3"))
    {
      cam_fdiv = 3;
    }
}


// image format callback

void
speed_val_cb(Fl_Choice *w, void *x)
{
  Fl_Menu_ *mw = (Fl_Menu_ *)w;
  const Fl_Menu_Item* m = mw->mvalue();

  if (!strcmp(m->label(), "60"))
    {
      cam_speed = S_60;
    }

  if (!strcmp(m->label(), "30"))
    {
      cam_speed = S_30;
    }

  if (!strcmp(m->label(), "15"))
    {
      cam_speed = S_15;
    }

  if (!strcmp(m->label(), "10"))
    {
      cam_speed = S_10;
    }
  if (!strcmp(m->label(), "7.5"))
    {
      cam_speed = S_7;
    }
  if (!strcmp(m->label(), "5"))
    {
      cam_speed = S_5;
    }
  if (!strcmp(m->label(), "3.75"))
    {
      cam_speed = S_3;
    }
}


// File loading and storage

static char filebuf[100];
static char *filedir = NULL;

char *
get_dirname(void)
{
  if (filedir) return NULL;   // use the last filename found

  filedir = getenv("DCAMDIR");
  if (filedir)
    {
      sprintf(filebuf, "%s/data", filedir); 
      filedir = filebuf;
    }
  else
    filedir = getcwd(NULL,0);
  return filedir;
}

void
store_current_cb(Fl_Menu_ *w, void *u)
{
  char nbuf[100];
  if (this_image == NULL)
    return;
  char *name = fl_file_chooser("Store current image to file", "{*.bmp}", get_dirname());

  double gam = 0.0;
  if (dcam && dcam->isBayer)
    gam = 0.60;

  if (name)
    {
      int n = strlen(name);
      if (n > 4 && !strcmp(&name[n-4], ".bmp"))
        name[n-4] = 0;
      strcpy(nbuf,name);
      strcat(nbuf,".bmp");
      if (im_color)
        write_color_bmp(nbuf, this_image, im_width, im_height, gam);
      else
        write_bmp(nbuf, this_image, im_width, im_height, gam);
    }
}


void
incr_name(char *name)
{
  int n = strlen(name);
  int d = name[n-1];
  if (d < '9')
    {
      d++;
      name[n-1] = d;
    }
  else
    {
      name[n-1] = '0';
      d = name[n-2];
      if (d < '9')
	{
	  d++;
	  name[n-2] = d;
	}     
      else
	{
	  name[n-2] = '0';
	  d = name[n-3];
	  if (d < '9')
	    d++;
	  else
	    d = '0';        // wrap around
	  name[n-3] = d;
	}     
    }
}


void
store_buffer_cb(Fl_Menu_ *w, void *u)
{
  char nbuf[100];
  if (this_image == NULL)
    return;
  char *name = fl_file_chooser("Store current image to file", "{*.bmp}", get_dirname());

  double gam = 0.0;
  if (dcam && dcam->isBayer)
    gam = 0.60;

  if (name)
    {
      // get down to base name
      int n = strlen(name);
      if (n > 4 && (!strcmp(&name[n-4], ".bmp")))
	{
	  name[n-4] = 0;
	  n -= 4;
	}
      char nn[1024];
      if (!(n > 3) ||		// check if we have an incrementing name
	  !isdigit(name[n-1]) ||
	  !isdigit(name[n-2]) ||
	  !isdigit(name[n-3]))
	{                   // no, make it so...
	  strcpy(nn,name);
	  name = nn;
	  name[n] = '0';
	  name[n+1] = '0';
	  name[n+2] = '0';
	  name[n+3] = 0;
	}

      int i;
      for (i=0; i<bufferNum; i++)
	{
	  if (bufferList[i])
	    {
	      strcpy(nbuf,name);
	      strcat(nbuf,".bmp");
	      if (im_color)
		write_color_bmp(nbuf, bufferList[i], im_width, im_height, gam);
	      else
		write_bmp(nbuf, bufferList[i], im_width, im_height, gam);
	      incr_name(name);	// increment
	    }
	}
      debugMessage("Wrote %d files at %s", i, name);
    }
}



//
// Buffer frames
//

void
show_frame_val(int n)
{
  mainw->frame_val->value(n);
  if (n >= mainw->frame_val->maximum())
    {
      mainw->record->value(false);
    }
}


void
show_buffer()
{
  if (acquire_type == AQT_BUFFER && bufferNum > 0)
    mainw->frame_val->activate();
  else
    mainw->frame_val->deactivate();

  if (acquire_type == AQT_CONTIN && bufferSize > 0)
    mainw->record->activate();
  else
    mainw->record->deactivate();

  if (bufferNum > 0)
    mainw->clear->activate();
  else
    mainw->clear->deactivate();

  mainw->frame_val->value(bufferCur);
  mainw->frame_val->maximum(bufferNum);
  //  mainw->buffer_num->val(1);
  //  mainw->buffer_num->val(bufferNum);
  //  mainw->buffer_num->maximum(bufferSize);
//  mainw->record->val(storeFlag);
}




// Store the current image set into the buffer

void do_storage()
{
  if (!storeFlag) return;
  if (acquire_type == AQT_CONTIN && this_image != NULL)
    {
      if (bufferCur+1 < bufferSize)
        {
	  if (bufferNum > 0)	// check for first frame
	    bufferCur++;
	  if (bufferList[bufferCur] == NULL)
	    bufferList[bufferCur] = new unsigned char[im_size];
	  memcpy(bufferList[bufferCur], this_image, im_size);
          bufferNum = bufferCur+1;
        }
      else
        {
          storeFlag = false;
	  mainw->record->value(false);
	}
    }
  show_buffer();
}


// Start up recording

void
record_cb(Fl_Light_Button *w, void *u)
{
  storeFlag = (bool)w->value();
  if (acquire_type != AQT_CONTIN)
    storeFlag = false;
  w->value(storeFlag);
  show_buffer();
}

// clear frames, leave buffers

void
clear_cb(Fl_Button *w, void *u)
{
  bufferCur = bufferNum = 0;
  mainw->frame_val->maximum(0);
  storeFlag = false;
  show_buffer();
}

// frame image change request

void
frame_val_cb(Fl_Counter *w, void *u)
{
  if (bufferNum < 0)
    return;
  bufferCur = (int)w->value() - 1;
  if (bufferCur < -1)
    bufferCur = bufferNum - 2;  // set to just before last frame
  else if (bufferCur < 0)
    bufferCur = bufferNum - 1;  // set to last frame
  single_shot = true;
}


// display frames

void
display_cb(Fl_Light_Button *w, void *x)
{
}


// frame rate

void
show_rate(double val)
{
  mainw->rate_val->val(val);
}

// Framegrabber type

void
camera_cb(class Fl_Widget *w, void *x)
{
  Fl_Menu_ *mw = (Fl_Menu_ *)w;
  const Fl_Menu_Item* m = mw->mvalue();
  int i = m->label()[0] - '0';
  cam_num = i;
}


//
// retrieve the next image and display it
// returns false if no image, i.e., not capturing
//

bool
get_next_image()
{
  static unsigned long lasttime = 0;
  if (!single_shot)		// are we getting frames?
    return false;

  if (acquire_type == AQT_NONE)	// not getting images
    return false;
  if (acquire_type == AQT_BUFFER) // get images from buffer
    {
      single_shot = false;
      if (bufferNum > 0)        // have something in the buffer
        {
          bufferCur++;          // increment current frame,
          if (bufferCur >= bufferNum) // check overflow
            bufferCur = 0;
          this_image = bufferList[bufferCur];
          show_buffer();        // display current frame
        }
      else
	return false;
    }
  else				// get images from camera
    {
      if (!dcam)		// no camera
	{
	  single_shot = false;
	  return false;
	}
      unsigned long time = 0;
      int frame;
      if (dcam->GetImage(&this_image, 500, &frame, &time))
	{
#if 0
	  char outp[256];
	  sprintf(outp, "%d: %d %d", (int)frame, (int)time, (int)(time - lasttime));
	  lasttime = time;
	  debugWin->Print(outp);
#endif
	}
      else
	return false;
    }

  // found a frame
  if (mainw->display_val->value())
    mainw->image->DrawImage(this_image, dcam);
  if (mainw->contin_val->value())
    single_shot = true;
  else
    single_shot = false;

  return true;
}


//
// top-level loop for grabbing and displaying
// 

void
top()
{
  int nframes = 0;
  int oldtime = ms_time();
  while (1)
    {
      if (!Fl::check())
        return;			// we're done here

      if (get_next_image())	// see if we can grab one and display it
	{
	  nframes++;
	  do_storage();		// see if we need to store it
	}
      else
        Sleep(100);

      if (nframes >= 10)	// display frame rate
        {
          show_rate(10000.0 / (double)(ms_time() - oldtime));
          oldtime = ms_time();
          nframes = 0;
        }
    }
}


//
// Set up camera menu items
//
void
set_cameras()
{
  int i;
  char mbname[100];
  Fl_Menu_Bar *mb = mainw->menubar; // get the menubar

  //  mb->clear();			// reset it
  for (i=0; i<dSys->NumCameras(); i++)
    {
      sprintf(mbname, "Camera/%d %s", i, dSys->Names()[i]);
      if (i == 0)
        mb->add(mbname, 0, camera_cb, (void *)NULL, FL_MENU_TOGGLE | FL_MENU_VALUE | FL_MENU_RADIO);
      else
        mb->add(mbname, 0, camera_cb, (void *)NULL, FL_MENU_TOGGLE | FL_MENU_RADIO);
    }
  cam_num = 0;
}




// Interface to top-level application call

#ifdef WIN32
#include <windows.h>
#include <windowsx.h>
#include "resource.h"

HINSTANCE hInstance;		/* application instance */

// returns an icon instance for the program
char *
get_icon(void)
{
  return (char *)LoadIcon(hInstance, MAKEINTRESOURCE(DCAM_ICON));
}

// OS-dependent message
void
print_msg(char *msg)
{
  MessageBox(NULL, msg, "Error", MB_OK);  
}

// reference to the main program
int main(int argc, char **argv);

// OS-dependent startup code
int WINAPI WinMain (HINSTANCE hInst, HINSTANCE hPrevInstance,
                    LPSTR lpszCmdLine, int nCmdShow)
{
  hInstance = hInst;
  main(0,NULL);
  return 0;
}

#else
// Linux definitions
char *get_icon(void)
{
  return NULL;
} 

void print_msg(char *msg)
{
  printf(msg);
}
#endif


// overlay drawing demo, just draws a red rectangle
void 
drawit(dcamWin *win, void *a)
{
  int x1 = win->width/4, y1 = win->height/4;
  int x2 = x1+win->width/2, y2 = y1+win->height/2;
  fl_color(FL_RED);
  fl_line(x1, y1, x2, y1);
  fl_line(x2, y1, x2, y2);
  fl_line(x2, y2, x1, y2);
  fl_line(x1, y2, x1, y1);
}


char *apparg = "dcam";

int 
main(int argc, char **argv)
{
  Fl::visual(FL_RGB);
  mainw = new fld();
  mainw->win->icon(get_icon());
  mainw->win->show(1,&apparg);	// need args to get icon correct
  mainw->rate_val->val("%3.1f"); // set float format
  mainw->rate_val->val(0.0);
  mainw->size_val->value(2);	// set to 640x480
  mainw->format_val->value(0);	// set to RGB24
  mainw->speed_val->value(2);	// set to 15 fps

  mainw->exposure_auto_button->value(true);
  mainw->exposure_val->deactivate();
  mainw->gain_val->deactivate();
  mainw->brightness_auto_button->value(true);
  mainw->brightness_val->deactivate();
  mainw->balance_auto_button->value(true);
  mainw->red_val->deactivate();
  mainw->blue_val->deactivate();
  mainw->imager_gamma_button->value(true);

  // set up debugging window for printing
  debugWin = new dcamDebugWin(100,100,500,400, "DCAM Debug Window");
  debugMessage("DCAM version %s", dcamVersion);
  dcamDebug = debugFn;

  dSysInit();
  if (dSys == NULL)		// oops, problems initializing
    {
      printf("Can't initialize IEEE 1394 bus\n");
      exit(-1);
    }
  else
    debugMessage("%d cameras available", dSys->NumCameras());

  // overlay demo drawing, set overlay draw callback
  if (drawOverlayFlag)
    mainw->image->DrawOverlayFn(drawit, NULL);

  // clear buffer
  int i;
  for (i=0; i<bufferNum; i++)
    bufferList[i] = NULL;

  set_cameras();

#ifdef DCAM_FW_DIALOG
  mainw->firmware_params_val = (Fl_Menu_Item *)mainw->menubar->find_item("Firmware...");
#endif

  top();
  return 0;
}
