//
// flwin.cpp
// 
// display window class, based on FLTK windows
//
//
// Copyright 2000 by Videre Design 
//


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "flwin.h"


// construct a double-buffered display window

dcamWin::dcamWin(int x, int y, int w, int h)
     : Fl_Double_Window(x,y,w,h)
{
  width = dwidth = w;
  height = dheight = h;
  xoff = yoff = 0;
  skipw = 1;
  linesize = w;
  pixelData = NULL;
  pixelType = S_MONO;
  end();
  bhandler = NULL;
  ovArg = NULL;
  ofn = NULL;
  displayGamma = 0.0;
  oldgamma = -1;
}

void
dcamWin::ClearImage()
{
  pixelData = NULL;

  //fl_rectf(xoff, yoff, width, height, 0, 0, 0);

  redraw();
}

//
// Interface function for drawing an image
//

void
dcamWin::DrawImage(unsigned char *im, dCamera *dc)
{
  int h = 120, w = 160, x = 0, y = 0;
  if (dc->Size() == S_1280x960)
    { w = 1280; h = 960; }
  if (dc->Size() == S_640x480)
    { w = 640; h = 480; }
  if (dc->Size() == S_320x240)
    { w = 320; h = 240; }
  if (dc->Size() == S_160x120)
    { w = 160; h = 120; }

  DisplayImage(im, w, h, w, dc->Type(), 1, displayGamma, x, y);
}



//
// image display function
// will try to decimate image to fit in display window
//

void
dcamWin::DisplayImage(unsigned char *im, int w, int h, int ls, int type, 
		      int s, double gam, int x, int y)
{
  // save these for redraws
  pixelData = im;
  pixelType = type;
  xoff = x;
  yoff = y;
  width = w;
  height = h;
  linesize = ls;
  skipw = s;

  while (height > dheight) 
  {
    height = height / 2;
    yoff = yoff / 2;
    linesize *= 2;
  }
  while (width > dwidth)
  {
    width = width / 2;
    skipw *= 2;
    //    linesize *= 2;
    xoff = xoff / 2;
  }

  if (gam != oldgamma)
    {
      oldgamma = gam;
      if (gam > 0.0 && gam < 1.0)
	{
	  for (int i=0; i<256; i++)
	    gamtab[i] = (int)(0.5 + 255.0 * pow((double)i/255.0, gam));
	}
    }

  if (!shown()) return;

  drawit();

#if 0
  make_current();
  if (pixelType == S_MONO || pixelType == S_YUV422)
    fl_draw_image_mono(pixelData, xoff, yoff, width, height, skipw, linesize);
  else if (pixelType == S_BAYER)
    //    fl_draw_image_bayer(pixelData, xoff, yoff, width, height, skipw, linesize);
    {}
  else if (pixelType == S_RGB24)
    fl_draw_image(pixelData, xoff, yoff, width, height, skipw*3, linesize*3);
#endif
}



struct image_cb_data
{
  unsigned char *pdata;
  int ndisp;
  int skip;
  int linesize;
  unsigned char *gtab;		// table of gamma values
};

// copy n pixels to out, starting at x,y
static void
disp_image_cb_fn(void *data, int x, int y, int n, uchar *out)
{
  int i, v, ndisp, skip;
  short *in;
  struct image_cb_data *cbd = (struct image_cb_data *)data;
  in = (short *)cbd->pdata;	// disparity images are short's
  in += y*cbd->linesize + x;
  ndisp = cbd->ndisp;
  skip = cbd->skip;
  for (i=0; i<n; i++, in+=skip)
    {
      v = *in;
      if (v < 0)
	*out++ = 0;
      else
	*out++ = v;
    }
};

// copy n pixels to out, starting at x,y
static void
gamma_image_cb_fn(void *data, int x, int y, int n, uchar *out)
{
  int i, ndisp, skip;
  unsigned char *in;
  unsigned char *gtab;
  struct image_cb_data *cbd = (struct image_cb_data *)data;
  in = (unsigned char *)cbd->pdata;	// mono image
  in += y*cbd->linesize + x;
  ndisp = cbd->ndisp;
  skip = cbd->skip;
  gtab = cbd->gtab;
  for (i=0; i<n; i++, in+=skip)
    *out++ = gtab[*in];
};


// copy n pixels to out, starting at x,y
static void
gamma_color_image_cb_fn(void *data, int x, int y, int n, uchar *out)
{
  int i, ndisp, skip;
  unsigned char *in;
  unsigned char *gtab;
  struct image_cb_data *cbd = (struct image_cb_data *)data;
  in = (unsigned char *)cbd->pdata;	// RGB24 image
  in += y*cbd->linesize + x;
  ndisp = cbd->ndisp;
  skip = cbd->skip;
  gtab = cbd->gtab;
  for (i=0; i<n; i++, in+=skip)
    {
      *out++ = gtab[*in];
      *out++ = gtab[*(in+1)];
      *out++ = gtab[*(in+2)];
    }
};


void
dcamWin::drawit()
{
  if (!shown()) return;

  struct image_cb_data cbd;
  make_current();

  if (pixelType == S_MONO)
    {
      if (displayGamma > 0.0 && displayGamma < 1.0)
	{
	  cbd.pdata = pixelData;
	  cbd.ndisp = 256;
	  cbd.linesize = linesize;
	  cbd.skip = skipw;
	  cbd.gtab = gamtab;
	  fl_draw_image_mono(gamma_image_cb_fn, (void *)&cbd, xoff, yoff, width, height, 1);
	}
      else
        fl_draw_image_mono(pixelData, xoff, yoff, width, height, skipw, linesize);
    }
  else if (pixelType == S_RGB24)
    {
      if (displayGamma > 0.0 && displayGamma < 1.0)
	{
	  cbd.pdata = pixelData;
	  cbd.ndisp = 256;
	  cbd.linesize = linesize*3;
	  cbd.skip = skipw*3;
	  cbd.gtab = gamtab;
	  fl_draw_image(gamma_color_image_cb_fn, (void *)&cbd, xoff, yoff, width, height, 3);
	}
      else
	fl_draw_image(pixelData, xoff, yoff, width, height, skipw*3, linesize*3);
    }
}



// re-drawing function, copies pixelData to the image
void
dcamWin::draw()
{
  if (height < dheight || width < dwidth || !pixelData)
  {
    fl_color(FL_BLACK);
    fl_rectf(0,0,dwidth,dheight);
  }
  if (pixelData)
    DisplayImage(pixelData, width, height, linesize, pixelType, skipw, xoff, yoff);
  if (ofn)
    (* ofn)(this, ovArg);
}			


// mouse event handling, let the user specify a callback via AddButtonHandler

int 
dcamWin::handle(int e) 
{
  int key, mod;
  int x, y;
  if (Fl_Window::handle(e)) return 1;
  mod = Fl::event_state();	// modifiers (shift, alt, etc)
  key = Fl::event_key();
  x = Fl::event_x();
  y = Fl::event_y();

  if (e == FL_PUSH && bhandler) 
    {
      if (Fl::focus() != this)
	Fl::focus(this);
      bhandler(x, y, (mod & 0x0F000000), (mod & 0x00FF0000), this);
      return 1;
    }
  if (e == FL_KEYBOARD)
    {
      // may want to add something here...
      return 1;
    }

  return 0;
}

// Fl_Value creation

Fl_Value::Fl_Value(int x, int y, int w, int h, const char *l, const char *val)
    : Fl_Widget(x,y,w,h,l) 
{ 
  strcpy(str,val); 
  strcpy(buf, val); 
  align(FL_ALIGN_LEFT); 
  labelcolor(FL_BLACK); 
  labelfont(FL_HELVETICA);  // we need this if the size is to change
  labelsize(14);
  c = FL_BLACK;
}

//
// debug window
//

//
// debug window
//

void copy_cb(Fl_Widget*, void* v) {
  dcamDebugWin* e = (dcamDebugWin*)v;

  if (!e->dbuf->selected()) return;
  const char *copy = e->dbuf->selection_text();
  if (*copy) Fl::copy(copy, strlen(copy), 1);
}

void save_cb(Fl_Widget*, void* v) {
  dcamDebugWin* e = (dcamDebugWin*)v;
  char *newfile;
  newfile = fl_file_chooser("Save File As?", "*.txt", "");
  if (newfile != NULL) 
    e->dbuf->savefile(newfile);
}


Fl_Menu_Item menuitems[] = {
  { "&File",              0, 0, 0, FL_SUBMENU },
    { "&Save File",       FL_CTRL + 's', (Fl_Callback *)save_cb },
  //    { "E&xit", FL_CTRL + 'q', (Fl_Callback *)quit_cb, 0 },
    { 0 },

  { "&Edit", 0, 0, 0, FL_SUBMENU },
    { "&Copy",       FL_CTRL + 'c', (Fl_Callback *)copy_cb },
    { 0 },

  { 0 }
};


dcamDebugWin::dcamDebugWin(int ix, int iy, int w, int h, char *name)
  : Fl_Window(ix,iy,w,h,name)
{
  begin();
  Fl_Menu_Bar* m = new Fl_Menu_Bar(0, 0, w, 30);
  m->copy(menuitems, this);
  dwin = new Fl_Text_Display(0,30,w,h);
  dbuf = new Fl_Text_Buffer;
  dwin->buffer(dbuf);
  dbuf->append("DCAM Debugging Window\n");
  dbuf->append("\n");
  resizable(dwin);
  end();
}


void
dcamDebugWin::Print(char *str, ...)
{
  if (str)
    {
      char buf[256];
      va_list ptr;
      va_start(ptr,str);
      vsprintf(buf,str,ptr);
      dbuf->append(buf);
      dbuf->append("\n");
      Bottom();
      va_end(ptr);
    }
}

void
dcamDebugWin::Bottom()
{
  dwin->insert_position(dbuf->length());
  dwin->show_insert_position();
}





