// 
// firmware dialog application for SVS
//
// Copyright 2001-2004 by Kurt Konolige, SRI International
// konolige@ai.sri.com
//
// The author hereby grants to SRI permission to use this software.
// The author also grants to SRI permission to distribute this software
// to schools for non-commercial educational use only.
//
// The author hereby grants to other individuals or organizations
// permission to use this software for non-commercial
// educational use only.  This software may not be distributed to others
// except by SRI, under the conditions above.
//
// Other than these cases, no part of this software may be used or
// distributed without written permission of the author.
//
// Neither the author nor SRI make any representations about the 
// suitability of this software for any purpose.  It is provided 
// "as is" without express or implied warranty.
//
// Kurt Konolige
// Senior Computer Scientist
// SRI International
// 333 Ravenswood Avenue
// Menlo Park, CA 94025
// E-mail:  konolige@ai.sri.com
//
//

//
// Generates a dialog for displaying and changing firmware parameters
//

#include "../src/dcam.h"

#ifdef WIN32
#include "direct.h"
#else
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#define Sleep(x) usleep(1000*x)
#include <ctype.h>
#endif

#include "fwdisp.h"
#include "FL/Fl_File_Chooser.H"
#include "FL/Fl_Browser.H"
#include "FL/x.H"
#include <FL/Fl_Menu_Bar.H>
#include <FL/fl_ask.H>

static firmware_disp *fwd = NULL;
static dCamera *dev = NULL;

static char devstr[5];
static int devid;
static int devtype;

static bool has_local_params;

// device types
#define DEV_STEREO 0x8
#define DEV_SINGLE 0x1
#define DEV_LEFT   0x3
#define DEV_RIGHT  0x5
#define DEV_REVERSED 0x10	// reversed image

/* Bus info block addr */
#define REG_ROM_BUS_INFO_BLOCK CONFIG_ROM_BASE+ROM_BUS_INFO_BLOCK
#define REG_ROM_ROOT_DIRECTORY CONFIG_ROM_BASE+ROM_ROOT_DIRECTORY
#define REG_ROM_UNIQUE_LEAF_ID CONFIG_ROM_BASE+UNIQUE_LEAF_ID
#define ROM_BUS_INFO_BLOCK             0x400U
#define ROM_ROOT_DIRECTORY             0x414U
#define UNIQUE_LEAF_ID                 0x428U 


//
// dc register definitions
//

#define CONFIG_ROM_BASE                0xF0000000UL
#define CCR_BASE                       0xF0F00000UL
#define LOCAL_BASE                     0xF0FF0000UL
#define PARAM1_OFFSET                  0x0400
#define CALIB_OFFSET                   0x0800
#define CALIB_SIZE                     0x1000           //  4 KB
#define DOWNLOAD_OFFSET                0x4000           
#define DOWNLOAD_SIZE                  0x3C00           // 15 KB
#define SCRATCHPAD_OFFSET              0x3C00
#define SCRATCHPAD_SIZE                0x80             // 128 B
#define LOCAL_PARAM_BASE               0xF0FFF000UL
#define CAM_STORE_MAGIC_NUM            0x5A01F687UL
#define FW_LEVEL_OFFSET                (1*4)
#define CAM_LEVEL_OFFSET               (2*4)
#define CAM_ID_OFFSET                  (3*4)
#define CAM_BLACK_OFFSET               (4*4)
#define CAM_START_ROW_OFFSET           (5*4)
#define CAM_START_COL_OFFSET           (6*4)
#define CAM_IMAGER_SIZE_OFFSET         (7*4)
#define CAM_FRAME_DIV_OFFSET           (8*4)
#define CAM_50HZ_OFFSET                (9*4)
#define CAM_PROC_OFFSET                (10*4)
#define CAM_PROC_THRESH_OFFSET         (11*4)


// we can enter bytes as integers

int ohci_crc16(unsigned int *data, int length)
{
        int check=0, i;
        int shift, sum, next=0;
	unsigned int d;

        for (i = length; i; i--) 
	  {
	    d = ((*data & 0xff) << 24) |
	        ((*data & 0xff00) << 8) |
	        ((*data & 0xff0000) >> 8) |
 	        ((*data & 0xff000000) >> 24);
	    for (next = check, shift = 28; shift >= 0; shift -= 4 ) 
	      {
		sum = ((next >> 12) ^ (d >> shift)) & 0xf;
		next = (next << 4) ^ (sum << 12) ^ (sum << 5) ^ (sum);
	      }
	    check = next & 0xffff;
	    data++;
	  }

        return ((check & 0xff) << 8) | ((check & 0xff00) >> 8);
}


void debugMessage(char *str, ...);

void
firmware_params_dialog(dCamera *dcam)
{
  if (dcam == NULL || dcam->initialized == false)
    {
      fl_alert("No device is opened");
      return;
    }

  dev = dcam;

  int fwsize = 380;
  if (fwd == NULL)
    fwd = new firmware_disp();
  else if (fwd->win->visible())
    fwsize = 532;

  // find local param block
  unsigned long qval;
  dev->ReadQuadlet(LOCAL_PARAM_BASE+FW_LEVEL_OFFSET, &qval);
  int major = (qval & 0x0000ff00)>>8;
  int minor = (qval & 0x000000ff);
  has_local_params = true;
  if ((qval >> 16) != 0 || major < 2 || minor > 10)	// check for local parameters
    has_local_params = false;

  // fw firmware version
  fwd->fw_version_val->val("%04x");
  fwd->fw_version_val->val(dev->fwFirmware);

  // cam firmware version
  fwd->camera_version_val->val("%04x");
  fwd->camera_version_val->val(dev->camFirmware);

  // color capability
  if (dev->has_color)
    fwd->color_cap_val->val("yes");
  else
    fwd->color_cap_val->val("no");

  // imager type
  switch(dev->camImager)
    {
    case CAM_TYPE_MT9M001:
      fwd->imager_type_val->val("MT9M001");      
      break;
    case CAM_TYPE_KAC1310:
      fwd->imager_type_val->val("KAC1310");      
      break;
    case CAM_TYPE_LM9648:
      fwd->imager_type_val->val("LM96x8");      
      break;
    case CAM_TYPE_MT9V403:
      fwd->imager_type_val->val("MT9V403");      
      break;
    case CAM_TYPE_MT9V022:
      fwd->imager_type_val->val("MT9V022");      
      break;
    default:
      fwd->imager_type_val->val("Unknown");      
      break;
    }

  // device type
  devtype = (dev->devID & 0xf0000) >> 16;
  switch (devtype)
    {
    case DEV_STEREO:
      fwd->device_type_val->value(1);
      break;
    case 0:
      fwd->device_type_val->value(0);
      break;
    case DEV_LEFT:
      fwd->device_type_val->value(2);
      break;
    case DEV_RIGHT:		// this shouldn't happen!
      fl_alert("Left camera of a pair is specified as right camera");
      fwd->device_type_val->value(2);
      break;
    case 0xFF:			// not yet set, but single cameras
      fwd->device_type_val->value(2);
      break;
    default:
      fwd->device_type_val->value(0);
      break;
    }

  // leave device type deactivated, since this is for
  //   stereo devices only
  fwd->device_type_val->set_output();
  fwd->swap_dev_val->deactivate();

  devid = dev->devID & 0xffff;
  sprintf(devstr, "%04x", devid);
  fwd->device_id_val->value(devstr);
  if (devid != 0xffff)		// id already present
    fwd->device_id_val->set_output();

  // local parameters
  if (!has_local_params)
    fwd->local_params_val->deactivate();
  else
    {
      // image size
      fwd->max_width_val->val("%4d");
      fwd->max_width_val->val(dev->max_width);
      fwd->max_height_val->val("%4d");
      fwd->max_height_val->val(dev->max_height);

      // inverted image
      if (devtype == DEV_STEREO)
        fwd->reversed_val->deactivate();
      else if (dev->isReversed)
	fwd->reversed_val->set();
      else
	fwd->reversed_val->clear();

      // 50 Hz operation
      if (!dev->has_50Hz_rate)
	fwd->Hz50_val->deactivate();
      else if (dev->is_50Hz_rate)
	fwd->Hz50_val->set();
      else
	fwd->Hz50_val->clear();	

      // Y offset
      if (!dev->has_y_offset)
	fwd->vert_off_val->deactivate();
      else 
	{
	  fwd->vert_off_val->bounds(0,dev->max_y_offset);
	  fwd->vert_off_val->value(dev->y_offset);
	}
    }

  // check for calibration params
  if (dev->GetCalib(NULL,0))	
    fwd->device_clear_cal_val->activate();
  else
    fwd->device_clear_cal_val->deactivate();

  fwd->win->size(359, fwsize);
  fwd->win->show();
}

//
// Device type callback
// Sets the device type, single-shot only 
//

void
device_type_cb(Fl_Choice *w, void *x)
{
  fwd->device_type_val->deactivate();
}

//
// Device id callback
// Sets the device id, single-shot only 
//

void
device_id_cb(Fl_Input *w, void *x)
{
  int val;
  unsigned long quad;
  unsigned long qbuf[10];
  int reversed = 0;

  char *id = (char *)w->value();
  bool ret = (bool)fl_ask("Are you sure you want to set the device STEREO #%s?", id);
  if (!ret)
    return;

  int serial = strtol(id, NULL, 16);

  // check if we can read the local param addresses
  unsigned long qval;
  int lp = 0xffffffff;
  ret = dev->ReadQuadlet(LOCAL_PARAM_BASE+CAM_LEVEL_OFFSET, &qval);
  debugMessage("[Device]  Camera type:  %08x", qval);
  if (ret)
    lp = qval;
  ret = dev->ReadQuadlet(LOCAL_PARAM_BASE+CAM_ID_OFFSET, &qval);
  printf("  Camera ID:    %08x\n", qval);

  if (lp != 0xffffffff)
    {
      debugMessage("  *** Setting ID in local block");
      qval = (reversed << 24) | serial;
      ret = dev->WriteQuadlet(LOCAL_PARAM_BASE+CAM_ID_OFFSET, qval);
      ret = dev->ReadQuadlet(LOCAL_PARAM_BASE+CAM_ID_OFFSET, &qval);
      debugMessage("  Camera ID:    %08x", qval);
      dev->SaveLocalParams();
    }
  else
    debugMessage("  Couldn't access local params");

  serial |= (DEV_STEREO)<<16;

  // set up ROM id
  dev->ReadQuadlet(REG_ROM_BUS_INFO_BLOCK, &quad);
  qbuf[0] = quad;
  debugMessage("  ROM bus info block first quad: %08x", quad);
  dev->ReadQuadlet(REG_ROM_BUS_INFO_BLOCK+16, &quad);
  debugMessage(  "  Chip ID low:                   %08x", quad);

  debugMessage("  Setting serial number to       %08x", serial);
  quad = serial;
  dev->WriteQuadlet(REG_ROM_BUS_INFO_BLOCK+16, quad);
  Sleep(100);
  dev->ReadQuadlet(REG_ROM_BUS_INFO_BLOCK+16, &quad);
  qbuf[4] = quad;
  debugMessage(  "  Chip ID low:                   %08x", quad);

  dev->ReadQuadlet(REG_ROM_BUS_INFO_BLOCK+4, &quad);
  qbuf[1] = quad;
  dev->ReadQuadlet(REG_ROM_BUS_INFO_BLOCK+8, &quad);
  qbuf[2] = quad;
  dev->ReadQuadlet(REG_ROM_BUS_INFO_BLOCK+12, &quad);
  qbuf[3] = quad;

  val = ohci_crc16((unsigned int *)&qbuf[2], 4);
  debugMessage("  Setting CRC to                     %04x", val);
  quad = val | (0xffff0000 & qbuf[0]);
  dev->WriteQuadlet(REG_ROM_BUS_INFO_BLOCK, quad);
  Sleep(100);
  dev->ReadQuadlet(REG_ROM_BUS_INFO_BLOCK, &quad);
  debugMessage("  ROM bus info block first quad: %08x", quad);

  // set unique leaf id
  dev->ReadQuadlet(REG_ROM_UNIQUE_LEAF_ID, &quad);
  qbuf[0] = quad;
  debugMessage("  ROM unique leaf ID first quad: %08x", (unsigned int)quad);
  dev->ReadQuadlet(REG_ROM_UNIQUE_LEAF_ID+8, &quad);
  debugMessage(  "  Chip ID low:                   %08x", (unsigned int)quad);

  debugMessage("  Setting serial number to       %08x", serial);
  quad = serial;        
  dev->WriteQuadlet(REG_ROM_UNIQUE_LEAF_ID+8, quad);
  Sleep(100);
  dev->ReadQuadlet(REG_ROM_UNIQUE_LEAF_ID+8, &quad);
  qbuf[2] = quad;
  debugMessage(  "  Chip ID low:                   %08x", (unsigned int)quad);

  dev->ReadQuadlet(REG_ROM_UNIQUE_LEAF_ID+4, &quad);
  qbuf[1] = quad;

  val = ohci_crc16((unsigned int *)&qbuf[1], 2);
  debugMessage("  Setting CRC to                     %04x", val);
  quad = val | (0xffff0000 & (unsigned int)qbuf[0]);
  dev->WriteQuadlet(REG_ROM_UNIQUE_LEAF_ID, quad);
  Sleep(100);
  dev->ReadQuadlet(REG_ROM_UNIQUE_LEAF_ID, &quad);
  debugMessage("  ROM unique leaf ID first quad: %08x", (unsigned int)quad);

  fwd->device_type_val->value(1); // stereo device

  fwd->device_id_val->deactivate(); // we're done
}


//
// local parameter callbacks

void
Hz50_cb(Fl_Check_Button *w, void *x)
{
  unsigned long qval;
  bool ret;
  if (w->value())		// set to 50 Hz
    {
      debugMessage("[Device] *** Setting 50 Hz rate");
      qval = 0x00010001;
    }
  else
    {
      debugMessage("[Device] *** Setting 60 Hz rate");
      qval = 0x00010000;
    }
  ret = dev->WriteQuadlet(LOCAL_PARAM_BASE+CAM_50HZ_OFFSET, qval);
  ret = dev->ReadQuadlet(LOCAL_PARAM_BASE+CAM_50HZ_OFFSET, &qval);
  debugMessage("[Device] Camera rate:    %08x", qval);
  fwd->device_save_val->activate();
}

// inverted images

void
reversed_cb(Fl_Check_Button *w, void *x)
{
  unsigned long qval;
  bool ret;
  if (w->value())		// set to 50 Hz
    {
      debugMessage("[Device] *** Setting inverted image on left imager");
      qval = 0x01000000 | devid;
      dev->isReversed = true;
    }
  else
    {
      debugMessage("[Device] *** Setting normal image on left imager");
      qval = 0x00000000 | devid;
      dev->isReversed = false;
    }
  ret = dev->WriteQuadlet(LOCAL_PARAM_BASE+CAM_ID_OFFSET, qval);
  ret = dev->ReadQuadlet(LOCAL_PARAM_BASE+CAM_ID_OFFSET, &qval);
  debugMessage("[Device] Camera id and inversion:    %08x", qval);
  fwd->device_save_val->activate();
}

void
reversed_right_cb(Fl_Check_Button *w, void *x)
{
}

// vertical offsets

void
vert_off_cb(Fl_Value_Slider *w, void *x)
{
  debugMessage("[Device] *** Setting vertical offset on imager");
  dev->SetYOffset((int)w->value());
  fwd->device_save_val->activate();
}

void
vert_off_right_cb(Fl_Value_Slider *w, void *x)
{
}

// save parameters to device
void
device_save_cb(Fl_Button *w, void *x)
{
  debugMessage("[Device] Uploading params to device");
  dev->SaveLocalParams();
  fwd->device_save_val->deactivate();
}

// clear calibration
void
device_clear_cal_cb(Fl_Button *w, void *x)
{
  bool ret = (bool)fl_ask("Are you sure you want to clear the calibration?");
  if (!ret)
    return;
  debugMessage("[Device] Clearing calibration on device");
  dev->SetCalib(NULL,0);	// clear it
  fwd->device_clear_cal_val->deactivate();
}



// swap devices (VAR models)
void
swap_dev_cb(Fl_Check_Button *w, void *x)
{
}


//
// imager register actions
//
static void
do_reg_read(int reg, int which)
{
  bool ret;
  int v1, v2;
  unsigned long qval;
  if (which > 8) 
    qval = (which << 24) | (reg << 8); // query 16 bit addr
  else
    qval = (which << 24) | (reg << 16); // query 8 bit addr

  ret = dev->WriteQuadlet(LOCAL_PARAM_BASE, qval);
  Sleep(10);
  ret = dev->ReadQuadlet(LOCAL_PARAM_BASE, &qval);
  v1 = (qval & 0xff00) >> 8;
  v2 = qval & 0xff;
  debugMessage("[Device] Register %02x: %02x %02x", reg, v1, v2);
  char regout[10];
  sprintf(regout, "%02x %02x", v1, v2);
  fwd->reg_value_val->value(regout);
}

void
reg_read_cb(Fl_Button *w, void *x)
{
  int n;
  unsigned int first, last;
  if (!dev) return;
  char *regstr;
  regstr = (char *)fwd->reg_addr_val->value();
  n = sscanf(regstr, "%x-%x", &first, &last);
  if (n == 2)			// matched a range!
    {
      if (first > last) 
	{
	  n = first;
	  first = last;
	  last = n;
	}
      for (n=first; n<=last; n++)
	do_reg_read(n,2);
    }
  else				// single return
    {
      int reg = strtol(regstr, NULL, 16);
      do_reg_read(reg,2);
    }
}


void
reg_write_cb(Fl_Button *w, void *x)
{
  if (!dev) return;
  char *regstr;
  regstr = (char *)fwd->reg_addr_val->value();
  int reg = strtol(regstr, NULL, 16);

  bool ret;
  int v, v1, v2;
  regstr = (char *)fwd->reg_value_val->value();
  v = strtol(regstr, NULL, 16);
  v = v & 0xffff;
  // write 16 bit result
  unsigned long qval = 0x04000000 | (reg << 16) | v; 
  ret = dev->WriteQuadlet(LOCAL_PARAM_BASE, qval);
  Sleep(10);

#if 0
  qval = 0x02000000 | (reg << 16);
  ret = dev->WriteQuadlet(LOCAL_PARAM_BASE, qval);
  svsSleep(10);
  ret = dev->ReadQuadlet(LOCAL_PARAM_BASE, &qval);
  v1 = (qval & 0xff00) >> 8;
  v2 = qval & 0xff;
  debugMessage("[Device] Register %02x: %02x %02x", reg, v1, v2);
  char regout[10];
  sprintf(regout, "%02x %02x", v1, v2);
  fwd->reg_value_val->value(regout);
#endif
}


void
reg_dump_cb(Fl_Button *w, void *x)
{
}


// 
// DCAM register read/write
//

void
dcam_reg_read_cb(Fl_Button *w, void *x)
{
  int n;
  unsigned int addr;
  if (!dev) return;
  char *regstr;
  regstr = (char *)fwd->reg_addr_val->value();
  addr = strtol(regstr, NULL, 16);
  bool ret;
  unsigned long qval;
  ret = dev->ReadQuadlet(CCR_BASE+addr, &qval);
  debugMessage("[Device] DCAM Register %02x: %04x", addr, qval);
  char regout[10];
  sprintf(regout, "%04x", qval);
  fwd->reg_value_val->value(regout);
}

