//
// image_io.cpp
//
// functions for saving BMPs
//
// Copyright 2000 by Videre Design 
//

#ifdef UNIX
#include <unistd.h>
#include <sys/types.h>
#define STRNCASECMP strncasecmp
#else
#include <io.h>
#include <fcntl.h>
#define STRNCASECMP _strnicmp
#endif
#include <stdio.h>
#include <malloc.h>

#include <math.h>

static int im_gamma[256];    // gamma correction array

int Print_error_messages = 1;
char *svsError = "";

#ifndef UNIX
int bcmp(char *x, char *y, int n) 
{
  int i;
  for (i=0; i<n; i++)
    if (*x++ != *y++) return 1;
  return 0;
}

void bzero(char *x, int n) 
{
  int i;
  for (i=0; i<n; i++)
    *x++ = 0;
}

void bcopy(char *x, char *y, int n) 
{
  int i;
  for (i=0; i<n; i++)
    *y++ = *x++;
}
#endif



/* BMP files */

int
read_bmp_word(unsigned char *p)
{
  int i;
  i = *p++;
  return ((*p << 8) | i);
}

long
read_bmp_long(unsigned char *p)
{
  int i, j, k;
  i = *p++;
  j = *p++;
  k = *p++;
  return ((*p << 24) | (k << 16) | (j << 8) | i);
}

#define BMP_HEADER_SIZE 14
#define BMP_INFO_SIZE 40
#define BMP_MAX_SIZE 1400000

char read_ret[100];

int
read_bmp(char *name, unsigned char *op, int *x, int *y, int interlace)
{
  unsigned char buf[2000];
  int offset, size, bpp, i, width, height;
  FILE *fp;
  int fd;

  if ((fp = fopen(name, "rb")) == NULL)
    {
      svsError = "Can't open file";
      return -1;
    }
  fd = fileno(fp);
#ifndef UNIX
  _setmode(fd,_O_BINARY);
#endif
  i = read(fd, buf, BMP_HEADER_SIZE+BMP_INFO_SIZE);
  if (buf[0] != 'B' || buf[1] != 'M')
    {
      svsError = "Not a BMP file";
      return -1;
    }
  offset = read_bmp_long(&buf[10]);
  size = read_bmp_long(&buf[2]);
  if (size > BMP_MAX_SIZE)
    {
	  sprintf(read_ret, "BMP file too large: size is %d bytes", size);
	  svsError = read_ret;
	  return -1;	
    }
  width = read_bmp_long(&buf[18]);
  height = read_bmp_long(&buf[22]);
  if (interlace) height = height/2;
  if (*x > 0 && width != *x)
    {
       svsError = "BMP file widths not consistent";
       return -1;
    }
  /* correct little problem with oddsize images */
  if (width == 640 && height > 480 && height < 490) height = 480;
  if (*y > 0 && height != *y)
    {
       svsError = "BMP file heights not consistent";
       return -1;
    }

  *x = width;
  *y = height;
  bpp = read_bmp_word(&buf[28]);
  if (bpp != 8)
    {
      svsError = "BMP file not 8 bits/pixel";
      return -1;
    }
  i = read(fd, buf, offset-BMP_INFO_SIZE-BMP_HEADER_SIZE); /* skip to data */
  op = op + width*(height-1);
  for (i=0; i<height; i++, op-=width)
    {
      if (interlace) 
	read(fd, op+(width*height), width);
      read(fd, op, width);
    }
  fclose( fp );
  return 0;
}



/* writing a BMP */

void
write_bmp_word(unsigned char *p, int v)
{
  *p++ = v & 0xff;
  *p = (v>>8) & 0xff;
}

void
write_bmp_long(unsigned char *p, int v)
{
  *p++ = v & 0xff;
  *p++ = (v>>8) & 0xff;
  *p++ = (v>>16) & 0xff;
  *p = (v>>24) & 0xff;
}

static unsigned char fbuf[4000];


int
write_bmp(char *name, unsigned char *op, int width, int height, double gam)
{
  unsigned char *p;
  int i;
  int fd, n;
  FILE *fp;

  if ((fp = fopen(name, "w")) == NULL)
    {
      svsError = "Can't open file for write";
      return -1;
    }
  fd = fileno(fp);
#ifndef UNIX
  _setmode(fd,_O_BINARY);
#endif
  fbuf[0] = 'B';
  fbuf[1] = 'M';
  write_bmp_long(&fbuf[2], width*height+BMP_HEADER_SIZE+BMP_INFO_SIZE
		 + 4*256);
  write_bmp_long(&fbuf[6], 0);
  write_bmp_long(&fbuf[10], BMP_HEADER_SIZE+BMP_INFO_SIZE+4*256);
  write_bmp_long(&fbuf[14], 40);	/* size of info header */
  write_bmp_long(&fbuf[18], width);
  write_bmp_long(&fbuf[22], height);
  write_bmp_word(&fbuf[26], 1);	/* number of planes */
  write_bmp_word(&fbuf[28], 8);	/* bits per pixel */
  write_bmp_long(&fbuf[30], 0);
  write_bmp_long(&fbuf[34], 0);
  write_bmp_long(&fbuf[38], 0);
  write_bmp_long(&fbuf[42], 0);
  write_bmp_long(&fbuf[46], 256);
  write_bmp_long(&fbuf[50], 256);
  p = fbuf + BMP_HEADER_SIZE + BMP_INFO_SIZE;

  if (gam > 0.0) 
    {		// gamma-corrected values
      for (int i=0; i<256; i++)
        im_gamma[i] = (int)(0.5 + 255.0 * pow((double)i/255.0, gam));
    }
  else
    for (int i=0; i<256; i++)
      im_gamma[i] = i;


  for (i=0; i<256; i++)
    {
      *p++ = im_gamma[i];
      *p++ = im_gamma[i];
      *p++ = im_gamma[i];
      *p++ = 0;
    }

  n = write(fd, fbuf, BMP_HEADER_SIZE+BMP_INFO_SIZE+4*256);
  op += width*(height-1);
  for (i=0; i<height; i++, op-=width)
    n = write(fd, (char *)op, width);
  fclose( fp );
  return 0;
}


//
// Write a color (24-bit) BMP
// 
// Each line ends of 4-byte (or perhaps 8-byte?) boundary
//
// Input data is 4 bytes/pixel
// Output data is packed, 3 bytes/pixel
//
// Total size of data is line size (line width * 3 rounded up to byte
// boundary) times height.
//
// Added ability to gamma-correct image
//



int
write_color_bmp(char *name, unsigned char *op, int width, int height, double gam)
{
  int i, j;
  int fd, n;
  FILE *fp;
  int linewidth;
  int ret;

  if ((fp = fopen(name, "wb")) == NULL)
    {
      svsError = "Can't open file for write";
      return -1;
    }
  fd = fileno(fp);
#ifndef UNIX
// _setmode(fd,_O_BINARY);
#endif

  linewidth = 3*width;
  linewidth = (linewidth + 0x3) & ~0x3;	/* round to 4 bytes */

  fbuf[0] = 'B';
  fbuf[1] = 'M';
  write_bmp_long(&fbuf[2], linewidth*height+BMP_HEADER_SIZE+BMP_INFO_SIZE);
				//total size
  write_bmp_long(&fbuf[6], 0);	// ??


  write_bmp_long(&fbuf[10], BMP_HEADER_SIZE+BMP_INFO_SIZE);
				// 0x36 (dec 54) for 24 bit pixels

  write_bmp_long(&fbuf[14], BMP_INFO_SIZE); // size of info header 
					     // 0x28 (dec 40) for 24 bit image

  write_bmp_long(&fbuf[18], width); // line width in pixels
  write_bmp_long(&fbuf[22], height); // number of lines


  write_bmp_word(&fbuf[26], 1);	// number of planes, 0x1 for 24 bit
  write_bmp_word(&fbuf[28], 24);// bits per pixel, 0x18 (dec 24) for 24 bit
  write_bmp_long(&fbuf[30], 0);  // compression
  write_bmp_long(&fbuf[34], linewidth*height);  // total size of data
  write_bmp_long(&fbuf[38], 0);  
  write_bmp_long(&fbuf[42], 0);  
  write_bmp_long(&fbuf[46], 0); 
  write_bmp_long(&fbuf[50], 0); 

  ret = write(fd, fbuf, BMP_HEADER_SIZE+BMP_INFO_SIZE);
  printf("%d", ret);
  op += 3*width*(height-1);

  if (gam > 0.0) 
    {		// gamma-corrected values
      for (int i=0; i<256; i++)
        im_gamma[i] = (int)(0.5 + 255.0 * pow((double)i/255.0, gam));
    }
  else
    for (int i=0; i<256; i++)
      im_gamma[i] = i;

  for (i=0; i<height; i++, op-=3*width)
    {
    for (j=0; j<width; j++)
	{
#if WIN32
          fbuf[j*3] = im_gamma[op[j*3+2]];	// BMPs are in B,G,R format; so is MS Windows file
	  fbuf[j*3+1] = im_gamma[op[j*3+1]];
	  fbuf[j*3+2] = im_gamma[op[j*3]];
#else
          fbuf[j*3] = im_gamma[op[j*3+2]];	// BMPs are in B,G,R format; so is MS Windows file
	  fbuf[j*3+1] = im_gamma[op[j*3+1]];
	  fbuf[j*3+2] = im_gamma[op[j*3]];
#endif
	}
      fbuf[j*3+3] = 0;
      fbuf[j*3+4] = 0;
      fbuf[j*3+5] = 0;
      n = write(fd, (char *)fbuf, linewidth);
    }
  fclose( fp );
  return 0;
}


  
